Imports DotNetNuke.Entities.Modules
Imports Aricie.DNN.Modules.LuceneSearch.ModuleProviders
Imports Aricie.DNN.Services
Imports Aricie.Services
Imports DotNetNuke.Services.Exceptions
Imports Aricie.DNN.Modules.LuceneSearch.Settings
Imports Aricie.DNN.Settings
Imports Lucene.Net.Documents
Imports Aricie.Collections
Imports Aricie.DNN.Modules.LuceneSearch.Logging
Imports DotNetNuke.Services.Search
Imports System.Reflection
Imports LuceneSearch.Monitoring.Monitors

Namespace Aricie.DNN.Modules.LuceneSearch.Business
    ''' <summary>
    ''' Indexer Provider For DotNetNuke using Lucene.Net
    ''' </summary>
    Public Class LuceneIndexer
        Inherits DotNetNuke.Services.Search.IndexingProvider


        Private _Logger As StepLogger

        Private Shared CurrentLuceneIndexer As New LuceneIndexer
        Private Shared Lock As New Object()
        Protected Friend Shared ThreadPickedUpForRunningIndexation As Boolean = False
        <ThreadStatic()> _
        Protected Friend Shared ThisThreadHandlesSearchIndexationAndStorage As Boolean = False

#Region "Overrides"

        Public Shared Function RetrieveSearchIndexItems(ByVal portalID As Integer) As SearchItemInfoCollection
            Return CurrentLuceneIndexer.GetSearchIndexItems(portalID)
        End Function

        ''' <summary>
        ''' Get the searchs Item all over the portal
        ''' </summary>
        ''' <param name="PortalID">PortalID to use to index the content</param>
        ''' <returns>SearchItemInfoCollection containing all searchiteminfo's modules</returns>
        Public Overrides Function GetSearchIndexItems(PortalID As Integer) As DotNetNuke.Services.Search.SearchItemInfoCollection

            Dim settings As LucenePortalSettings = GetModuleSettings(Of LucenePortalSettings)(SettingsScope.PortalSettings, PortalID)
            _Logger = New StepLogger(settings.LogIndexingProcess, PortalID)

            Dim results As New SearchItemInfoCollection

            If (Not ThreadPickedUpForRunningIndexation) Then
                SyncLock Lock
                    If (Not ThreadPickedUpForRunningIndexation) Then
                        ThreadPickedUpForRunningIndexation = True
                        ThisThreadHandlesSearchIndexationAndStorage = True
                    End If
                End SyncLock
            End If

            If ThisThreadHandlesSearchIndexationAndStorage Then
                Try

                    _Logger.LogStep(LuceneLogType.Retrieval, "Thread " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString() + " starting indexation")

                    If Not String.IsNullOrEmpty(Settings.IndexFolderMapPath) Then

                        Dim CurrentIndexationMonitor As New IndexationMonitor

                        Dim nativeModuleNamesAlreadyDone As New Dictionary(Of String, Boolean)

                        results = GetNativeItems(PortalID, nativeModuleNamesAlreadyDone, CurrentIndexationMonitor)
                        _Logger.LogStep(LuceneLogType.Retrieval, "Native items retrieved", "Nb Search items", CurrentIndexationMonitor.NativeSearchItemsHandler.Count)

                        GetAdditionalNativeItems(PortalID, results, nativeModuleNamesAlreadyDone, CurrentIndexationMonitor)
                        _Logger.LogStep(LuceneLogType.Retrieval, "Extended module items retrieved", "Nb Search items", CurrentIndexationMonitor.ExtendedSearchItemsHandler.Count)

                        GetProgressiveStandAloneItems(PortalID, results, nativeModuleNamesAlreadyDone, CurrentIndexationMonitor)
                        _Logger.LogStep(LuceneLogType.Retrieval, "Progressive stand alone items retrieved", "Nb Search items", CurrentIndexationMonitor.ProgressiveStandaloneSearchItemsHandler.Count)

                        Dim luceneDataStore As LuceneSearchDataStore = DirectCast(SearchDataStoreProvider.Instance, LuceneSearchDataStore)
                        If luceneDataStore IsNot Nothing Then
                            Dim CurrentlyExistingIndexStructure As Components.Lucene.ExistingIndex.ExistingIndexStructure(Of Components.Lucene.ExistingIndex.ExistingIndexDocumentSmartOrdering) = Nothing
                            Using existingIndex As LuceneQueryResults = luceneDataStore.GetCurrentIndex(PortalID)
                                CurrentlyExistingIndexStructure = New Components.Lucene.ExistingIndex.ExistingIndexStructure(Of Components.Lucene.ExistingIndex.ExistingIndexDocumentSmartOrdering)(existingIndex)
                            End Using
                            ' we upgrade our items
                            results = Me.UpgradeSearchItems(PortalID, results, settings, CurrentlyExistingIndexStructure)

                            AddStandAloneItems(results, settings, PortalID, CurrentIndexationMonitor)
                            _Logger.LogStep(LuceneLogType.Retrieval, "Stand alone items retrieved", "Nb Search items", CurrentIndexationMonitor.StandaloneSearchItemsHandler.Count)
                        End If

                        GetReflexiveItems(results, CurrentIndexationMonitor)
                        _Logger.LogStep(LuceneLogType.Retrieval, "Reflexive items retrieved", "Nb Search items", CurrentIndexationMonitor.ReflexiveSearchItemsHandler.Count)
                        ' details statistiques
                        If settings.LogIndexingProcess Then
                            Dim MonitorLog As New Global.LuceneSearch.Monitoring.Logging.LogInfoProviderLogger(PortalID)
                            MonitorLog.LogInfo(CurrentIndexationMonitor)
                        End If
                    End If
                Catch ex As Exception
                    _Logger.LogStep(LuceneLogType.Retrieval, "Thread " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString() + " encountered an exception during indexation. Please see next entry for exception details")
                    Exceptions.LogException(ex)
                    ' une erreur est survenue, on abandonne la tache d'indexation  quelqu'un d'autre que nous
                    ThisThreadHandlesSearchIndexationAndStorage = False
                    ThreadPickedUpForRunningIndexation = False
                    ' on relance la mme exception
                    Throw
                End Try

                _Logger.LogLastStep(LuceneLogType.Retrieval, "Thread " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString() + " retrieved " + results.Count.ToString + " items")
            Else
                _Logger.LogLastStep(LuceneLogType.Retrieval, "Thread " + System.Threading.Thread.CurrentThread.ManagedThreadId.ToString() + " didn't pick up the indexation task because it is already handled by another thread")
            End If

            Return results
        End Function


#End Region


#Region "Indexation methods"
        Private Function GetNativeItems(ByVal portalId As Integer, ByRef nativeModuleNamesAlreadyDone As Dictionary(Of String, Boolean), CurrentIndexationMonitor As IndexationMonitor) As SearchItemInfoCollection
            Dim settings As LucenePortalSettings = GetModuleSettings(Of LucenePortalSettings)(SettingsScope.PortalSettings, portalId)
            Dim nativeItems As New SearchItemInfoCollection
            Dim tempBool As Boolean
            Dim currentISearchable As ISearchable


            For Each searchModule As SearchableElement In GetModuleList(portalId)
                If settings.IsAvailableForIndexing(searchModule.modInfo.ModuleName) Then

                    Dim moduleProvider As LuceneModuleProviderConfig = ProviderFactory.Instance.GetSearchableFromModuleName(searchModule.modInfo.ModuleName)
                    Dim upgradeModuleProvider As LuceneModuleProviderConfig = ProviderFactory.Instance.GetUpgraderFromModuleName(searchModule.modInfo.ModuleName)
                    Dim ModuleProviderName As String = LuceneSearchItemInfo.NO_PROVIDER
                    Dim UpgradeProviderName As String = String.Empty

                    If moduleProvider IsNot Nothing Then
                        If Not nativeModuleNamesAlreadyDone.TryGetValue(searchModule.modInfo.ModuleName, tempBool) Then
                            nativeModuleNamesAlreadyDone(searchModule.modInfo.ModuleName) = True
                        End If
                        currentISearchable = New CompoundISearchable(searchModule.ModControllerType, moduleProvider)
                        ModuleProviderName = moduleProvider.Name
                    Else
                        currentISearchable = searchModule.ModControllerType
                    End If

                    If upgradeModuleProvider IsNot Nothing Then
                        UpgradeProviderName = upgradeModuleProvider.Name
                    End If

                    Try
                        Dim toAdd As SearchItemInfoCollection = currentISearchable.GetSearchItems(searchModule.modInfo)

                        If toAdd IsNot Nothing Then
                            CurrentIndexationMonitor.NativeSearchItemsHandler.AddItems(searchModule.modInfo.ModuleName, searchModule.modInfo.ModuleID, toAdd.Count)

                            For Each objSearchItem As SearchItemInfo In toAdd
                                Dim lsii As LuceneSearchItemInfo = LuceneSearchItemInfo.GetLuceneSearchItem(objSearchItem, portalId, ModuleProviderName)
                                lsii.UpgradeProviderName = UpgradeProviderName
                                nativeItems.Add(lsii)
                            Next
                        End If
                    Catch ex As Exception
                        _Logger.LogStep(LuceneLogType.Retrieval, String.Format("Exception in module {0}: see next log for exception", searchModule.modInfo.ModuleName))
                        ExceptionHelper.LogException(ex)
                    End Try
                End If
            Next
            Return nativeItems
        End Function

        Private Sub GetAdditionalNativeItems(ByVal portalId As Integer, ByRef nativeItems As SearchItemInfoCollection, ByVal nativeModuleNamesAlreadyDone As Dictionary(Of String, Boolean), CurrentIndexationMonitor As IndexationMonitor)
            Dim settings As LucenePortalSettings = GetModuleSettings(Of LucenePortalSettings)(SettingsScope.PortalSettings, portalId)
            Dim currentISearchable As ISearchable
            For Each moduleProvider As LuceneModuleProviderConfig In ProviderFactory.Instance.GetSearchables
                If Not nativeModuleNamesAlreadyDone.ContainsKey(moduleProvider.ModuleName) Then
                    If settings.IsAvailableForIndexing(moduleProvider.ModuleName) Then
                        Try
                            '_Logger.LogStep(LuceneLogType.IndexRetrieval, String.Format("Additional native search: retrieving items for provider {0}", moduleProvider.ModuleName))

                            Dim UpgradeProvider As LuceneModuleProviderConfig = ProviderFactory.Instance.GetUpgraderFromModuleName(moduleProvider.ModuleName)
                            Dim UpgradeProviderName As String = String.Empty
                            If (UpgradeProvider IsNot Nothing) Then
                                UpgradeProviderName = UpgradeProvider.Name
                            End If

                            currentISearchable = moduleProvider.AsSearchable
                            Dim friendlyName As String = NukeHelper.ModuleDefinitionController.GetModuleDefinition(NukeHelper.GetModuleDefIdByModuleName(moduleProvider.ModuleName)).FriendlyName
                            For Each objModule As ModuleInfo In NukeHelper.ModuleController.GetModulesByDefinition(portalId, friendlyName)
                                Dim toAdd As SearchItemInfoCollection = currentISearchable.GetSearchItems(objModule)

                                If toAdd IsNot Nothing Then
                                    CurrentIndexationMonitor.ExtendedSearchItemsHandler.AddItems(objModule.ModuleName, objModule.ModuleID, toAdd.Count)

                                    For Each objSearchItem As SearchItemInfo In toAdd
                                        Dim lsii As LuceneSearchItemInfo = LuceneSearchItemInfo.GetLuceneSearchItem(objSearchItem, portalId, moduleProvider.Name)
                                        lsii.UpgradeProviderName = UpgradeProviderName
                                        nativeItems.Add(lsii)
                                    Next

                                End If
                            Next
                        Catch ex As Exception
                            ExceptionHelper.LogException(ex)
                        End Try
                    End If
                End If
            Next
        End Sub

        Private Sub GetProgressiveStandAloneItems(ByVal portalId As Integer, ByRef nativeItems As SearchItemInfoCollection, ByVal nativeModuleNamesAlreadyDone As Dictionary(Of String, Boolean), CurrentIndexationMonitor As IndexationMonitor)
            For Each ProviderConfig As LuceneProviderConfig In ProviderFactory.Instance.GetProgressiveStandAlonesProviders()
                If Not nativeModuleNamesAlreadyDone.ContainsKey(ProviderConfig.Name) Then
                    nativeModuleNamesAlreadyDone.Add(ProviderConfig.Name, True)
                    Try
                        Dim UpgradeProvider As LuceneProviderConfig = ProviderFactory.Instance.GetProviderByName(ProviderConfig.Name)
                        Dim UpgradeProviderName As String = String.Empty
                        If (UpgradeProvider IsNot Nothing) Then
                            UpgradeProviderName = UpgradeProvider.Name
                        End If

                        Dim currentISearchable As ILuceneStandAloneSearchable = DirectCast(ProviderConfig.Provider, ILuceneStandAloneProgressiveSearchable)

                        Dim toAdd As SearchItemInfoCollection = currentISearchable.GetSearchItems(portalId)

                        If toAdd IsNot Nothing Then
                            CurrentIndexationMonitor.ProgressiveStandaloneSearchItemsHandler.AddItems(ProviderConfig.Name, toAdd.Count)

                            For Each objSearchItem As SearchItemInfo In toAdd
                                Dim lsii As LuceneSearchItemInfo = LuceneSearchItemInfo.GetLuceneSearchItem(objSearchItem, portalId, ProviderConfig.Name)
                                lsii.UpgradeProviderName = UpgradeProviderName
                                nativeItems.Add(lsii)
                            Next

                        End If
                    Catch ex As Exception
                        ExceptionHelper.LogException(ex)
                    End Try
                End If
            Next
        End Sub

        Private Sub AddStandAloneItems(ByRef moduleSearchItems As SearchItemInfoCollection, ByVal settings As LucenePortalSettings, ByVal portalId As Integer, CurrentIndexationMonitor As IndexationMonitor)
            'indexing stand alones providers
            Dim nbItems As Integer = 0

            For Each standAloneProvider As LuceneProviderConfig In ProviderFactory.Instance.GetStandAlonesProviders()

                Try
                    Dim currentIStandalone As ILuceneStandAloneSearchable = standAloneProvider.AsStandaloneSearchable
                    Dim searchItems As SearchItemInfoCollection = currentIStandalone.GetSearchItems(portalId)
                    If searchItems IsNot Nothing Then
                        CurrentIndexationMonitor.StandaloneSearchItemsHandler.AddItems(standAloneProvider.Name, searchItems.Count)
                        For Each searchItem As SearchItemInfo In searchItems
                            Dim luceneItem As LuceneSearchItemInfo = LuceneSearchItemInfo.GetLuceneSearchItem(searchItem, portalId, standAloneProvider.Name)
                            luceneItem.ProviderName = standAloneProvider.Name
                            moduleSearchItems.Add(luceneItem)
                            nbItems += 1
                        Next

                    End If

                Catch ex As Exception
                    DotNetNuke.Services.Exceptions.Exceptions.LogException(ex)
                End Try

            Next
        End Sub

        Private Sub GetReflexiveItems(searchItems As SearchItemInfoCollection, CurrentIndexationMonitor As IndexationMonitor)
            Dim AdditionalReflexiveItems As New List(Of LuceneSearchItemInfo)
            Dim PreviousCount = 0
            For Each rp In ProviderFactory.Instance.ReflexiveProviders
                AdditionalReflexiveItems.AddRange(rp.EnrichSearchItems(searchItems))
                CurrentIndexationMonitor.ReflexiveSearchItemsHandler.AddItems(rp.GetType().Name, AdditionalReflexiveItems.Count - PreviousCount)
                PreviousCount = AdditionalReflexiveItems.Count ' cheating to keep track of the items added individually by the providers
            Next
            searchItems.AddRange(AdditionalReflexiveItems.ToArray())
        End Sub


#End Region

#Region "Upgrade methods"
        Private Function BeginUpgrade(p As IUpgradeEventsHandlerProvider, PortalId As Integer, Settings As LucenePortalSettings) As Integer
            'VB doesn't like lambdas that don't return values so we use a function
            p.BeginUpgrade(New UpgradeEventsHandlerProviderArguments With {.PortalId = PortalId, .Settings = Settings})
            Return 0
        End Function

        Private Function EndUpgrade(p As IUpgradeEventsHandlerProvider, PortalId As Integer, Settings As LucenePortalSettings) As Integer
            ' VB doesn't like lambdas that don't return values so we use a function
            p.EndUpgrade(New UpgradeEventsHandlerProviderArguments With {.PortalId = PortalId, .Settings = Settings})
            Return 0
        End Function

        Private Function UpgradeSearchItems(ByVal portalID As Integer, ByVal nativeItems As SearchItemInfoCollection, ByVal settings As LucenePortalSettings, ByVal existingIndex As Components.Lucene.ExistingIndex.ExistingIndexStructure(Of Components.Lucene.ExistingIndex.ExistingIndexDocumentSmartOrdering)) As SearchItemInfoCollection
            Dim UpgradedSearchItems As New SearchItemInfoCollection

            _Logger.LogStep(LuceneLogType.Upgrade, "Upgrade beginning")

            ' we warn providers that wew are beginning the upgrade process
            Dim ListeUpgradeEventHandlerProviders = ProviderFactory.Instance.GetUpgradeEventHandlerProviders()
            ListeUpgradeEventHandlerProviders.ForEach(Function(p) BeginUpgrade(p, portalID, settings))

            ' sorting items according to their priority relative to the current index
            Dim sortedNativeItems() As LuceneSearchItemInfo = nativeItems.ToArray().Cast(Of LuceneSearchItemInfo)().ToArray()
            Array.Sort(sortedNativeItems, New Components.Lucene.IndexMerger.IndexMergeSorter(existingIndex) With {.OldDocumentsTurnOver = settings.OldDocIndexingRatio})

            Dim CurrentUpgradeMonitor As New UpgradeMonitor(settings)

            For Each searchItem As LuceneSearchItemInfo In sortedNativeItems
                Dim MatchingIndexedDocument As Components.Lucene.ExistingIndex.ExistingIndexDocumentSmartOrdering = Nothing
                existingIndex.UidToDocument.TryGetValue(searchItem.UniqueId, MatchingIndexedDocument)
                ' we skip the document that we want to update if
                ' - its publication date is considered trustworthy and it has a previously indexed version where the publication date is near (below one second) the current publication date
                ' - we have used up all the allocated time to upgrade documents
                If ((settings.TrustPubDates AndAlso MatchingIndexedDocument IsNot Nothing AndAlso (searchItem.PubDate - MatchingIndexedDocument.PublicationDate).TotalSeconds < 1) _
                    OrElse CurrentUpgradeMonitor.UpgradeTimeExceeded) _
                Then
                    searchItem.IndexingStatus = IndexingItemStatus.Skipped
                    CurrentUpgradeMonitor.ItemsSkipped += 1
                Else
                    ' this document needs to be upgraded, either it's new or publication dates didn't mark it as too near its previous version
                    Dim searchProviderConfig As LuceneProviderConfig = ProviderFactory.Instance.GetProviderByName(searchItem.UpgradeProviderName)
                    If (searchProviderConfig IsNot Nothing AndAlso TypeOf searchProviderConfig.Provider Is ILuceneSearchableUpgrade) Then
                        CurrentUpgradeMonitor.UpgradeWatch.Start()
                        Try
                            DirectCast(searchProviderConfig.Provider, ILuceneSearchableUpgrade).UpgradeSearchItem(portalID, searchItem)
                            searchItem.ProviderName = searchProviderConfig.Name
                            CurrentUpgradeMonitor.ItemsUpgraded += 1
                        Catch ex As Exception
                            ExceptionHelper.LogException(ex)
                        End Try
                        CurrentUpgradeMonitor.UpgradeWatch.Stop()
                        If CurrentUpgradeMonitor.UpgradeTimeExceeded Then
                            ' our time window for upgrade is over
                            _Logger.LogStep(LuceneLogType.Upgrade, String.Format("Document upgrade exceeded timeout value ({0}s). Remaining documents will be processed on next indexing session", settings.NbSecondsUpgrading))
                        End If
                    Else
                        CurrentUpgradeMonitor.ItemsIgnored += 1
                    End If

                End If

                UpgradedSearchItems.Add(searchItem)
            Next
            CurrentUpgradeMonitor.UpgradeWatch.Start()
            nativeItems.Clear() ' reclaim some memory

            If (settings.LogIndexingProcess) Then
                Dim MonitorLog As New Global.LuceneSearch.Monitoring.Logging.LogInfoProviderLogger(portalID)
                MonitorLog.LogInfo(CurrentUpgradeMonitor)
            End If

            ' we warn providers that we are done with the upgrade
            ProviderFactory.Instance.GetUpgradeEventHandlerProviders().ForEach(Function(p) EndUpgrade(p, portalID, settings))

            _Logger.LogLastStep(LuceneLogType.Upgrade, "Upgrade ending")

            Return UpgradedSearchItems
        End Function

#End Region

#Region "Public Methods"

        Public Shared Function GetSearchableControllers() As Dictionary(Of String, ISearchable)

            Dim toReturn As New Dictionary(Of String, ISearchable)

            Dim controller As Object
            For Each desktopModule As DesktopModuleInfo In NukeHelper.DesktopModuleController.GetDesktopModules()
                If (desktopModule.BusinessControllerClass <> "") Then
                    Try
                        controller = ReflectionHelper.CreateObject(desktopModule.BusinessControllerClass)
                    Catch ex As Exception
                        ExceptionHelper.LogException(ex)
                        controller = Nothing
                    End Try

                    If controller IsNot Nothing AndAlso TypeOf controller Is ISearchable Then
                        toReturn(desktopModule.ModuleName) = DirectCast(controller, ISearchable)
                    End If
                End If
            Next
            Return toReturn
        End Function


        Public Shared Function GetSearchableDesktopModules(ByVal portalId As Integer) As List(Of DesktopModuleInfo)

            Dim toReturn As New List(Of DesktopModuleInfo)
            Dim controller As Object

            For Each desktopModule As DesktopModuleInfo In _
                NukeHelper.DesktopModuleController.GetDesktopModulesByPortal(portalId)
                If ProviderFactory.Instance.GetSearchableFromModuleName(desktopModule.ModuleName) IsNot Nothing Then
                    toReturn.Add(desktopModule)
                ElseIf (desktopModule.BusinessControllerClass <> "") Then
                    Try
                        controller = ReflectionHelper.CreateObject(desktopModule.BusinessControllerClass)
                    Catch ex As Exception
                        ExceptionHelper.LogException(ex)
                        controller = Nothing
                    End Try

                    ' test modifi pour DNN 7.1
                    If controller IsNot Nothing AndAlso _
                        ( _
                        TypeOf controller Is ISearchable _
                        OrElse _
                        controller.GetType().GetMethod("GetModifiedSearchDocuments") IsNot Nothing
                        ) _
                    Then
                        toReturn.Add(desktopModule)
                    End If
                End If
            Next
            Return toReturn
        End Function


#End Region


#Region "Private methods"

        Protected Class SearchableElement
            Public Property ModControllerType As ISearchable
            Public Property modInfo As ModuleInfo
        End Class

        ''' <summary>
        ''' Mthode rcupre depuis ModuleIndexer pour viter la dpendance directe sur une classe qui a chang de lieu dans DNN7
        ''' </summary>
        ''' <param name="PortalID">Identifiant portail</param>
        ''' <returns>Liste de ISearchable</returns>
        ''' <remarks></remarks>
        Protected Function GetModuleList(PortalID As Integer) As IEnumerable(Of SearchableElement)
            Dim MC As ModuleController = New ModuleController()
            Dim DMC As New DesktopModuleController()

            Dim BusinessClassToController As New Hashtable
            Dim AlreadyProcessedModules As New List(Of Integer)

            Dim resultList As New List(Of SearchableElement)

            For Each moduleInfo As ModuleInfo In MC.GetSearchModules(PortalID)
                If Not AlreadyProcessedModules.Contains(moduleInfo.ModuleID) Then
                    Dim DMI As DesktopModuleInfo = Nothing
                    Try
                        DMI = DMC.GetDesktopModule(moduleInfo.DesktopModuleID)
                        If Not String.IsNullOrEmpty(DMI.BusinessControllerClass) Then
                            Dim BusinessController As Object = BusinessClassToController(DMI.BusinessControllerClass)
                            If BusinessController Is Nothing Then
                                BusinessController = ReflectionHelper.CreateObject(DMI.BusinessControllerClass)
                                BusinessClassToController(DMI.BusinessControllerClass) = BusinessController
                            End If
                            If TypeOf BusinessController Is ISearchable Then
                                resultList.Add(New SearchableElement With {.modInfo = moduleInfo, .ModControllerType = DirectCast(BusinessController, ISearchable)})
                            End If
                        End If
                    Catch ex As Exception
                        Try
                            Dim message As String = String.Format("Error Creating BusinessControllerClass '{0}' of module({1}) id=({2}) in tab({3}) and portal({4}) ", DMI.BusinessControllerClass, DMI.ModuleName, moduleInfo.ModuleID, moduleInfo.TabID, moduleInfo.PortalID)
                            Throw New Exception(message, ex)
                        Catch exc As Exception
                            Exceptions.LogException(exc)
                        End Try
                    Finally
                        AlreadyProcessedModules.Add(moduleInfo.ModuleID)
                    End Try
                End If
            Next

            Return resultList
        End Function




        


#End Region


    End Class
End Namespace


